# -*- coding: utf-8 -*-
"""CONM_20250319_00.ipynb

Automatically generated by Colab.

Original file is located at
    https://colab.research.google.com/drive/19RiUkIgZlyDru1wkiejVdvVcNxxuID0V

# CondaManager
Conda 仮想環境におけるCUDA コンポーネントなどの管理を行うクラス

```
2025/03/19 0.1.0 BranchRouter 用ダミー生成
2025/03/19 0.1.1 デバッグ
```

# CondaManager

## 1. モジュール定義
"""

# %%writefile /content/drive/MyDrive/code/CondaManager.py
# @title a. CondaManager 定義
import os
import sys
import subprocess
import shutil
# from CondaEnvManager import CondaEnvManager
get_ipython().system( "pip install an_DebugHelper" )
from an_debughelper import DebugHelper
get_ipython().system( "pip install an_EnvManager" )
from an_envmanager import EnvManager
get_ipython().system( "pip install an_CudaUtility" )
from an_cudautility import CudaUtility


class CondaManager:
    """Conda 仮想環境における CUDA コンポーネントの管理を行うクラス"""
    def __init__(self):
        """CondaManager の初期化。デバッグ用オブジェクトを受け取る。"""
        self.debug = DebugHelper(instance_name="CondaManager")
        self.debug.enable_log_to_file_stdout()
        self.debug.enable_log_to_file_stderr()
        self.debug.enable_timestamp()
        if shutil.which("conda") is None:
            self.debug.log_step("Conda is not available. Ensure Conda is installed and activated.", success=False)
            raise Exception("Conda is not available. Ensure Conda is installed and activated.")
        # self.condaenv = CondaEnvManager()
        self.cudau = CudaUtility()
        self.envmanager = EnvManager(env_files=["settings_ven.env"])

    def cleanup_cuda(self, env_name = None ):
        """CUDA 関連のすべてのパッケージ (cudatoolkit, cudnn, cublas) をアンインストールする。"""
        # CUDA 関連パッケージのアンインストール開始をログに記録
        self.debug.log_step( "引数reset_cudaがTrueなので、CUDAのリセットを行います", success = True )
        self.debug.log_step("CUDA 関連パッケージ (cudatoolkit, cudnn, cublas) をアンインストールします。")
        try:
            # conda コマンドで CUDA パッケージを削除
            # conda remove -n kohya_env --yes cudatoolkit cudnn cublas
            # result, stdout, stderr = self.debug.run_command(f"conda remove -n {env} --yes cudatoolkit cudnn cublas", stdout = True, stderr = True)
            venv_folder = Path( self.envmanager.get_env_var( "Ven_venv_folder" ) )
            # envs = self.condaenv.list_envs()
            if not env_name is None:
                envs = [ env_name ]
            else:
                self.debug.log_step( "リセットする仮想環境が設定されていません", success = False )
                return

            for env in envs:
                self.debug.log_step( f"env: { env }", success = None )
                self.remove_cuda_packages_from_env( venv_folder.joinpath( env ) )
                # アンインストールが成功したことをログに記録
            self.debug.log_step("指定された仮想環境のCUDA 関連パッケージをすべて環境から削除しました。", success = True)
        except subprocess.CalledProcessError as e:
            # コマンド実行に失敗した場合はエラーをログに記録し、例外を再送出
            self.debug.log_step(f"CUDA パッケージの削除に失敗しました: {e}", success = False)
            raise Exception("CUDA パッケージの削除に失敗しました")



    def remove_cuda_packages_from_env( self, venv_folder ):
        """
        指定した環境 (env_path) から cudatoolkit, cudnn, cublas を削除する。
        存在しない場合はスキップする。
        """
        packages = ["cudatoolkit", "cudnn", "cublas"]
        for pkg in packages:
            # 環境内に対象パッケージがインストールされているか確認する
            cmd_check = f"conda list -p { venv_folder } | grep {pkg}"
            result, stdout, stderr = self.debug.run_command( command = cmd_check, stdout = True, stderr = True )
            self.debug.log_step( f"stdout: { stdout }", success = None )
            self.debug.log_step( f"stderr: { stderr }", success = None )
            if result == 0:
                self.debug.log_step(f"{pkg} が {venv_folder} に存在します。削除を試みます。", success=None)
                cmd_remove = f"conda remove -p {venv_folder} --yes {pkg}"
                try:
                    subprocess.run(cmd_remove, shell=True, check=True)
                    self.debug.log_step(f"{pkg} の削除に成功しました。", success=True)
                except subprocess.CalledProcessError as e:
                    self.debug.log_step(f"{pkg} の削除に失敗しました: {e}", success=False)
                    # pipでの削除は試みない
            elif result == 1:
                self.debug.log_step(f"{pkg} は {venv_folder} に存在しないため、スキップします。", success=True)
            else:
                self.debug.log_step("原因不明のエラーが発生しました", success=False)



# # 例: 仮想環境ディレクトリのベースパスと環境名のリスト
# base_env_dir = "/content/drive/MyDrive/cEnv/venv"
# envs = ["kohya_env", "sd_env", "comfy_env"]

# for env in envs:
#     full_env_path = os.path.join(base_env_dir, env)
#     print(f"\n=== 環境 {env} のCUDA関連パッケージを削除 ===")
#     remove_cuda_packages_from_env(full_env_path)


    def install_cuda( self, env_name = "kohya_env", cuda_version = "CU118" ):
        try:
            result, stdout, stderr = self.debug.run_command("conda list cudatoolkit", stdout=True, stderr=True)
            print( result )
            print( stdout )
            print( stderr )
        except subprocess.CalledProcessError as e:
            self.debug.log_step(f"Error checking cudatoolkit: {e}", success=False)
            raise Exception("Error checking cudatoolkit")
        # if self.cudau.is_valid_cuda_version(stdout):
        #     installed_cuda_version_symbol = self.cudau.version_symbol(self.extract_cudatoolkit_version(stdout))
        #     self.cuda_version = cuda_version
        #     self.debug.log_step(f"extract_cudatoolkit_version: {installed_cuda_version_symbol}", success=None)
        #     self.debug.log_step(f"               cuda_version: {cuda_version}", success=None)
        #     if installed_cuda_version_symbol == cuda_version:
        #         self.debug.log_step(f"CUDA Toolkit {cuda_version} is already installed. Skipping installation.", success=True)
        #         return
        sys.exit(0)

        self.debug.log_step( f"仮想環境: { env_name } をリセットし、CUDAを { cuda_version } にセットします。", success = None )
        self.cleanup_cuda( env_name = env_name )
        self.install_cuda_toolkit( cuda_version = cuda_version )
        self.install_cuda_dependencies( cuda_version = cuda_version )
        self.install_pytorch( cuda_version = cuda_version )
        # self.install_xformers()
        # self.install_bitsandbytes( cuda_version = cuda_version )
        self.set_cuda_environment_vars()
        # self.verify_cuda_installation() 環境チェックをするプログラムをつくったが、時間がかかるのでパス。
        self.debug.log_step( f"仮想環境: { env_name } がリセットされ、CUDAが { cuda_version } にセットされました。", success = True )

    def install_cuda_toolkit(self, cuda_version="CU118"):
        self.debug.log_step(f"Checking for CUDA Toolkit {self.cudau.version_number(cuda_version)} in the Conda environment...", success=None)
        # try:
        #     result, stdout, stderr = self.debug.run_command("conda list cudatoolkit", stdout=True, stderr=True)
        # except subprocess.CalledProcessError as e:
        #     self.debug.log_step(f"Error checking cudatoolkit: {e}", success=False)
        #     raise Exception("Error checking cudatoolkit")
        # if self.cudau.is_valid_cuda_version(stdout):
        #     installed_cuda_version_symbol = self.cudau.version_symbol(self.extract_cudatoolkit_version(stdout))
        #     self.cuda_version = cuda_version
        #     self.debug.log_step(f"extract_cudatoolkit_version: {installed_cuda_version_symbol}", success=None)
        #     self.debug.log_step(f"               cuda_version: {cuda_version}", success=None)
        #     if installed_cuda_version_symbol == cuda_version:
        #         self.debug.log_step(f"CUDA Toolkit {cuda_version} is already installed. Skipping installation.", success=True)
        #         return
        # mambaを使ってインストール
        cmd = f"mamba install -y cudatoolkit={self.cudau.version_number(cuda_version)} -c nvidia -c conda-forge"
        result, stdout, stderr = self.debug.run_command(command=cmd, stdout=True, stderr=True)
        self.debug.log_step(f"CUDA Toolkit {cuda_version} installation completed.", success=True)

    def extract_cudatoolkit_version(self, stdout: str) -> str:
        """
        conda list cudatoolkit の stdout から cudatoolkit のバージョン番号を抽出して返却する。

        例えば、stdout に以下のような行が含まれている場合:

        cudatoolkit               11.8.0               hd77b12b_0    conda-forge

        このメソッドは "11.8.0" を返却します。

        引数:
        stdout (str): "conda list cudatoolkit" の標準出力

        戻り値:
        cudatoolkit のバージョン文字列 (例："11.8.0") が抽出できた場合、その文字列。抽出できなかった場合は None を返す。
        """
        import re
        # 出力を行ごとに分割
        lines = stdout.strip().splitlines()
        # 各行を解析
        for line in lines:
            # ヘッダー行を除外し、"cudatoolkit" で始まる行を探す
            if line.lstrip().startswith("cudatoolkit"):
                # 正規表現でバージョン番号を抽出する
                m = re.search(r'^cudatoolkit\s+([0-9]+(?:\.[0-9]+)+)', line)
                if m:
                    return m.group(1)
        return None

    def install_cuda_dependencies(self, cuda_version="CU118"):
        if cuda_version not in ("CU118", "CU124"):
            self.debug.log_step(f"サポートされていない cuda_version '{cuda_version}' が指定されました。", success=False)
            raise ValueError("cuda_version は 'CU118' または 'CU124' のみ指定可能です。")
        if cuda_version == "CU118":
            commands = [
                "mamba install -y cudatoolkit=11.8 -c nvidia -c conda-forge",
                "mamba install -y cudnn=8.4.1.50 -c conda-forge"
            ]
        else:  # cuda_version == "CU124"
            commands = [
                "mamba install -y cudatoolkit=12.4 -c nvidia -c conda-forge",
                "mamba install -y cudnn=8.9 -c conda-forge"
            ]
        self.debug.log_step(f"CUDA バージョン {cuda_version} 用の依存パッケージをインストールします。", success=None)
        for cmd in commands:
            result, stdout, stderr = self.debug.run_command(command=cmd, stdout=True, stderr=True)
            self.debug.debug_print(f"result:\n {result}")
            self.debug.debug_print(f"stdout:\n {stdout}")
            self.debug.debug_print(f"stderr:\n {stderr}")
            if result != 0:
                self.debug.log_step(f"CUDA 依存パッケージのインストールに失敗しました: {stderr}", success=False)
                raise Exception("CUDA 依存パッケージのインストールに失敗しました")
        self.debug.log_step("CUDA 関連の依存パッケージのインストールが正常に完了しました。", success=True)

    def install_pytorch(self, cuda_version = "CU118" ):
        """PyTorch をインストールする。既にインストールされている場合はスキップする。"""
        self.debug.log_step("Checking for PyTorch in the Conda environment...", success=None)
        # cmd = f"conda install pytorch torchvision torchaudio cudatoolkit={ self.cudau.version_number( cuda_version ) } -c pytorch -c nvidia"
        cmd = f"mamba install pytorch torchvision torchaudio cudatoolkit={ self.cudau.version_number( cuda_version ) } -c pytorch -c nvidia"
        reset, stdout, stderr = self.debug.run_command( command = cmd, stdout = True, stderr = True, sync = False )
        self.debug.log_step(f"stdout:\n{stdout}", success = None)
        self.debug.log_step(f"stderr:\n{stderr}", success = None)
        if reset == 0:
            self.debug.log_step("PyTorch is already installed.", success=True)
        else:
            raise Exception("PyTorch のインストールに失敗しました")

    def install_xformers(self):
        """xformers パッケージを仮想環境にインストールする。"""
        self.debug.log_step("xformers をインストールします...", success=None)
        # 仮想環境内の python のパスを取得（例: /content/drive/MyDrive/cEnv/venv/kohya_env/bin/python）
        python_path = os.path.join(self.envmanager.get_env_var("Ven_venv_folder"), "kohya_env", "bin", "python")
        cmd = f"{python_path} -m pip install xformers"
        self.debug.log_step(f"インストールコマンド: {cmd}", success=None)
        result, stdout, stderr = self.debug.run_command(command=cmd, stdout=True, stderr=True)
        self.debug.log_step(f"xformers インストール stdout:\n{stdout}", success=None)
        self.debug.log_step(f"xformers インストール stderr:\n{stderr}", success=None)
        if result == 0:
            self.debug.log_step("xformers のインストールに成功しました。", success=True)
        else:
            raise Exception("xformers のインストールに失敗しました。")

    def install_bitsandbytes( self, cuda_version = "CU118" ):
        self.debug.log_step("Checking for Bitsandbytes in the Conda environment...", success=None)
        cuda_version_number = float( self.cudau.version_number( cuda_version) )
        if not (11.0 <= cuda_version_number <= 12.8):
            raise Exception("サポートされていない cuda_version が指定されました。")
        cmd = f"pip install bitsandbytes"
        reset, stdout, stderr = self.debug.run_command( command = cmd, stdout = True, stderr = True, sync = False )
        self.debug.log_step(f"stdout:\n{stdout}", success = None)
        self.debug.log_step(f"stderr:\n{stderr}", success = None)
        if reset == 0:
            self.debug.log_step("Bitsandbytes is already installed.", success=True)
        else:
            raise Exception("Bitsandbytes のインストールに失敗しました")

    def set_cuda_environment_vars(self):
        """CUDA_HOME 環境変数を設定し、LD_LIBRARY_PATH に CUDA ライブラリパスを追加する。"""
        # CUDA_HOME 環境変数を設定
        cuda_path = "/usr/local/cuda"
        os.environ["CUDA_HOME"] = cuda_path
        self.debug.log_step(f"環境変数 CUDA_HOME を {cuda_path} に設定しました。", success = True )
        # LD_LIBRARY_PATH 環境変数に CUDA のライブラリパスを追加
        cuda_lib_path = "/usr/local/cuda/lib"
        current_ld_path = os.environ.get("LD_LIBRARY_PATH", "")
        if cuda_lib_path not in current_ld_path.split(":"):
            new_ld_path = cuda_lib_path if current_ld_path == "" else f"{current_ld_path}:{cuda_lib_path}"
            os.environ["LD_LIBRARY_PATH"] = new_ld_path
        self.debug.log_step(f"環境変数 LD_LIBRARY_PATH に {cuda_lib_path} を追加しました。", success = True)

    def verify_cuda_installation(self):
        """CUDA 関連パッケージのインストール状況と PyTorch による GPU 認識を検証する。"""
        # インストール済みの CUDA 関連パッケージを確認
        # packages = ["cudatoolkit", "cudnn", "cublas"]
        packages = ["cudatoolkit", "cudnn" ]
        all_installed = True
        for pkg in packages:
            result, stdout, stderr = self.debug.run_command(f"conda list {pkg}", stdout = True, stderr = True)
            output = stdout
            if pkg not in output:
                self.debug.log_step(f"{pkg} が環境にインストールされていません。", success = False)
                all_installed = False
            else:
                # パッケージ行（名前とバージョン）をログに記録
                for line in output.splitlines():
                    if line.strip().startswith(pkg):
                        self.debug.log_step(f"{pkg} インストール済み: {line.strip()}")
                        break
        if not all_installed:
            raise RuntimeError("一部のCUDA関連パッケージが正しくインストールされていません。")
        # PyTorch による GPU 認識を確認  'print(\"Hello World!\")'
        cmd = f"{os.path.join( self.envmanager.get_env_var( 'Ven_venv_folder' ), 'kohya_env', 'bin', 'python')} -c 'import torch; print(torch.cuda.is_available())'"
        self.debug.log_step( f"cmd : { cmd }", success = None )
        result, stdout, stderr = self.debug.run_command( command = cmd , stdout = True, stderr= True )
        self.debug.log_step( f"stdout: { stdout }", success = True )
        self.debug.log_step( f"stderr: { stderr }", success = False )
        torch_check = stdout.strip()
        self.debug.log_step(f"torch.cuda.is_available() の結果: {torch_check}")
        if result == 0:
            if torch_check != "True":
                self.debug.log_step("PyTorch が CUDA 対応GPUを認識できませんでした。", success = False )
                raise RuntimeError("PyTorch が GPU を認識できません。CUDA のインストールに問題がある可能性があります。")
            else:
                self.debug.log_step("CUDA 環境は正常です (PyTorch が GPU を認識しました)。", success = True )
        else:
            raise Exception( f"resultが0以外の値を返しています{ result }")